"""(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Dec 5, 2021
"""
import copy
from itertools import combinations
from abc import ABC, abstractmethod
import numpy as np
import pandas as pd
from causallib.positivity import BasePositivity


class MultipleTreatmentPositivity(BasePositivity, ABC):
    """
    Abstract class defining the methods require for a multiple treatment positivity scheme
    """

    def __init__(self, base_positivity_estimator, verbose=False):
        """

        Args:
            base_positivity_estimator: (MLHLS positivity model): The positivity model to be used consecutively for the
            positivity calculations.
            verbose: If true a progress bar will appear while fitting the single estimators
        """
        self.base_positivity_estimator = base_positivity_estimator
        self.positivity_estimators_ = dict()
        self.verbose = verbose

    def predict(self, X, a=None):
        """
            Predicts which rows in X are positive and which are not
        Args:
            X (pd.DataFrame): Covariates to predict on, each row is a instance and column is a feature
            a (pd.Series): The treatment assignment vector
        Returns:
            (pd.Series): A boolean series that indicate which instances are overlapping (True value) and which is
            not overlapping (False value)
        """
        return self.positivity_profile(X, a).min(axis=1).astype(bool)

    def _calculate_single_positivity(self, rows: pd.Series, treatment: str, X: pd.DataFrame, a=None) -> pd.Series:
        """
            Calculates the positivity indication on the covariates in the specified rows
        Args:
            rows: The rows to predict the covariates on
            treatment: The name of the single treatment for positivity prediction, (must correspond to a key in the
                        positivity_estimators)
            X (pd.DataFrame): Covariates to predict on, each row is a instance and column is a feature
            a: The treatment of the corresponding covariates, transferred to the base positivity estimator (for example
                matching requires the treatment)

        Returns:
            (pd.Series) : Boolean series of rows length indicating whether subject is within positivity overlap
        """
        a_tmp = None if a is None else a[rows]
        return (
            self.positivity_estimators_[treatment]
                .predict(X.loc[rows, :], a=a_tmp)
                .values
        )

    @abstractmethod
    def _create_positivity_profile_columns(self):
        """
        Calculates the columns of the positivity profile
        Returns:
            columns for the positivity profile
        """

    def positivity_profile(
            self, X: pd.DataFrame, a=None, estimation_population='Both'
    ) -> pd.DataFrame:
        """
            Creates the positivity profile for the covariates in X for certain populations
        Args:
            X (pd.DataFrame): Covariates to predict on, each row is a instance and column is a feature.
            a (pd.Series): The treatments given to the patients.
            estimation_population (str): Can be "treated" (treated group only), "control" (control group only), "both"
            (either treated or control group) or Every (every patient that was given a treatment) . If a is not supplied
            then the mode every is selected automatically.

        Returns:

        """
        positivity_profile = pd.DataFrame(
            columns=self._create_positivity_profile_columns(), index=X.index,
            dtype="boolean",
        )
        for treatment in positivity_profile.columns:
            rows = self._get_treated_rows_indicator(X, a, estimation_population, treatment=treatment)
            positivity_profile.loc[rows, treatment] = self._calculate_single_positivity(X=X, a=a, rows=rows,
                                                                                        treatment=treatment)
        return positivity_profile

    @abstractmethod
    def _treatment_positivity_summary_table(
            self, X: pd.DataFrame, a=None, estimation_population='Both'
    ) -> pd.DataFrame:
        """
        Returns the positivity treatment profile, the rate of positivity per treatment
        Args:
            X (pd.DataFrame): Covariates to predict positivity on
            a (pd.Series): Treatment assignment series
            estimation_population (str): Can be treated (treated group only), control (control group only) both
            (either treated or control group) or Every (every patient that was given a treatment) . If a is not supplied
            then the mode every is selected automatically.

        Returns:
            pd.DataFrame: A table of the rate of treatments.
        """

    @abstractmethod
    def _treatment_positivity_summary_contingency(
            self, X: pd.DataFrame, a=None, estimation_population='Both'
    ) -> pd.DataFrame:
        """
        The method returns the positivity profile in a form of contingency matrix where every treatment
        appears once in the rows and once in the columns of the profile. This form might be easier for viewing
        in the one versus another case and can be used for graph based algorithms. The values in the mat are rate of
        overlapping treatments by both single estimations.
        Args:
            X (pd.DataFrame): Covariates to predict on
            a (pd.Series) : The treatment assignment to be passed to the base positivity if required and
            corresponding single positivity estimator or all of the patients'
            estimation_population (str): Can be treated (treated group only), control (control group only) both
            (either treated or control group) or Every (every patient that was given a treatment) . If a is not supplied
            then the mode every is selected automatically.
        Returns:
            pd.DataFrame: A contingency table of the rate of treatments.

        """

    def treatment_positivity_summary(
            self, X: pd.DataFrame, a=None, estimation_population='Both', as_contingency_table=False
    ) -> pd.DataFrame:
        """
            The method summarized the positivity rate per treatment and reference group
        Args:
            X (pd.DataFrame): Covariates to predict on
            a (pd.Series) : The treatment assignment to be passed to the base positivity if required and
            estimation_population (str): Can be treated (treated group only), control (control group only) both
            (either treated or control group) or Every (every patient that was given a treatment) . If a is not supplied
            then the mode every is selected automatically.
            as_contingency_table (bool): If true returns a symmetric with every treatment as row and columns.
            otherwise, return a matrox where the rows correspond to the first treatment of the pair and the columns
            to the second treatment in each treatment pair. This form might be easier for viewing
            in the one versus another case and can be used for graph based algorithms.
        Returns:
            pd.DataFrame|pd.Series : a data frame describing the positivity rate of each single positivity estimation
        """
        if as_contingency_table:
            return self._treatment_positivity_summary_contingency(X=X, a=a, estimation_population=estimation_population)
        else:
            return self._treatment_positivity_summary_table(X=X, a=a, estimation_population=estimation_population)

    @abstractmethod
    def _get_treated_rows_indicator(self, X, a, estimation_population='Both', treatment=None):
        """
        Returns indicator variable stating which row in x is to be used for the positivity prediction.
        If the treatment vector is not supplied every coordinate will be used for positivity estimation
        Args:
            X (pd.DataFrame): Covariates
            a (pd.Series) : Treatment assignments of the treated
            estimation_population (str): Can be treated (treated group only), control (control group only) both
            (either treated or control group) or Every (every patient that was given a treatment) . If a is not supplied
            then the mode every is selected automatically.

        Returns:

        """


class OneVersusRestPositivity(MultipleTreatmentPositivity):
    """
    This is class converts the multiple treatment positivity calculation to
    a multiple positivity calculations The class receives a multi-treatment
    data and a base positivity estimator and creates multiple estimators one
    per treatment where the one treatment is the treated and the rest are the control.
    """

    def fit(self, X, a):
        """
            Fits a one versus the rest positivity in it the experimental class is the treated and every other treatment
            class is the control jointly
        Args:
            X (pd.DataFrame): covariates to predict on, each row is a instance and column is a feature
            a (pd.Series): A series of length equal to X with the treatment assignment

        Returns:
            one_versus_all_positivity: A causal model with an inner models fitted.
        """
        clf = self.base_positivity_estimator
        if self.verbose:
            print("Fitting {} treatment".format(len(a.unique())))
        for treatment in a.unique():
            clf = copy.deepcopy(clf)
            clf.fit(X, a == treatment)
            self.positivity_estimators_[treatment] = clf
        return self

    def _create_positivity_profile_columns(self):
        """
        Create a list with the names of the treatment as the columns
        Returns:
            a set containing the names of the treatments to be used as columns
        """
        kys = list(self.positivity_estimators_.keys())
        kys.sort()
        return kys

    def _treatment_positivity_summary_table(
            self, X: pd.DataFrame, a=None, estimation_population='Both'
    ) -> pd.DataFrame:
        """
        Returns the positivity treatment profile, the rate of positivity per treatment
        Args:
            X (pd.DataFrame): Covariates to predict positivity on
            a (pd.Series): Treatment assignment series
            estimation_population (str): Can be treated (treated group only), control (control group only) both
            (either treated or control group) or Every (every patient that was given a treatment) . If a is not supplied
            then the mode every is selected automatically.

        Returns:
            pd.DataFrame: A table of the rate of treatments.
        """
        return self.positivity_profile(X=X, a=a, estimation_population=estimation_population).mean(
            axis=0)

    def _treatment_positivity_summary_contingency(
            self, X: pd.DataFrame, a=None, estimation_population='Both'
    ) -> pd.DataFrame:
        """
        The method returns the positivity profile in a form of contingency matrix where every treatment
        appears once in the rows and once in the columns of the profile. This form might be easier for viewing
        in the one versus another case and can be used for graph based algorithms. The values in the mat are rate of
        overlapping treatments by both single estimations.
        Args:
            X (pd.DataFrame): Covariates to predict on
            a (pd.Series) : The treatment assignment to be passed to the base positivity if required and
            corresponding single positivity estimator or all of the patients'
            estimation_population (str): Can be treated (treated group only), control (control group only) both
            (either treated or control group) or Every (every patient that was given a treatment) . If a is not supplied
            then the mode every is selected automatically.
        Returns:
            pd.DataFrame: A contingency table of the rate of treatments.

        """
        treatment_profile = self.positivity_profile(X=X, a=a, estimation_population=estimation_population)
        treats = treatment_profile.columns.sort_values().values
        contingency_frame = pd.DataFrame(index=treats, columns=treats)
        for a1, a2 in combinations(treats, 2):
            contingency_frame.loc[a1, a2] = np.mean(
                np.logical_and(
                    treatment_profile[a1].fillna(False),
                    treatment_profile[a2].fillna(False)
                )
            )
        return contingency_frame.astype(float)

    def _get_treated_rows_indicator(self, X, a, estimation_population='Both', treatment=None):
        """
        Returns indicator variable stating which row in x is to be used for the positivity prediction.
        If the treatment vector is not supplied every coordinate will be used for positivity estimation
        Args:
            X (pd.DataFrame): Covariates
            a (pd.Series) : Treatment assignments of the treated
            estimation_population (str): Can be treated (treated group only), control (control group only) both
            (either treated or control group) or Every (every patient that was given a treatment) . If a is not supplied
            then the mode every is selected automatically.

        Returns:

        """
        if a is None or estimation_population.lower() in ['every', 'both']:
            return pd.Series(np.repeat(True, X.shape[0]), index=X.index)
        elif estimation_population.lower() == 'control':
            return a != treatment
        elif estimation_population.lower() == 'treated':
            return a == treatment


class OneVersusAnotherPositivity(MultipleTreatmentPositivity):
    """
    This is class converts the multiple treatment positivity calculation to
    multiple one vs. one positivity calculations. The class receives a multi-treatment
    data and a base positivity estimator and creates multiple estimators one
    per treatment pair where the one treatment is the treated and the second is the control.
    """

    def __init__(
            self,
            base_positivity_estimator,
            verbose=False,
            treatment_pairs_list=None,
            treatment_a_name="treatmentA",
            treatment_b_name="treatmentB",
    ):
        """

        Args:
            base_positivity_estimator: (MLHLS positivity model): The positivity model to be used consecutively for the
            positivity calculations.
            verbose: If true a progress bar will appear while fitting the single estimator
            treatment_pairs_list (None| tuple list|int|float|bool|str): A list of tuples each containing a pair
            of treatments to estimate the positivity between them. The first element is considered the control and the
            second is treated. If None a list of  all pairwise of treatments will be created.
            If the value is a string, bool or numeric it will be used the control treatment and a list  containing all
            pairs of treatment versus it will be created.
            treatment_a_name: name for the first enry of the data frame report multi index
            treatment_b_name: name for the second enry of the data frame report multi index
        """
        self.treatment_pairs_list = treatment_pairs_list
        self.treatment_a_name = treatment_a_name
        self.treatment_b_name = treatment_b_name
        super().__init__(
            base_positivity_estimator=base_positivity_estimator, verbose=verbose
        )

    def fit(self, X: pd.DataFrame, a: pd.Series) -> BasePositivity:
        """
            Fits a one versus the another positivity in it the user defines the pairs of treatment where one will be
            control and the other the experimental.
        Args:
            X (pd.DataFrame): covariates to predict on, each row is a instance and column is a feature
            a (pd.Series): A series of length equal to X with the treatment assignment
        Returns:
            one_versus_another_positivity: A MultipleTreatmentPositivity model with an inner models fitted.
        """
        if self.treatment_pairs_list is None:
            self.treatment_pairs_list = list(combinations(a.unique(), 2))
        elif isinstance(self.treatment_pairs_list, (int, float, bool, str)):
            self.treatment_pairs_list = [(v, self.treatment_pairs_list) for v in a.unique()
                                         if v != self.treatment_pairs_list]

        clf = self.base_positivity_estimator
        if self.verbose:
            print("Fitting treatment {}".format(len(self.treatment_pairs_list)))
        for a1, a2 in self.treatment_pairs_list:
            clf = copy.deepcopy(clf)
            clf.fit(X.loc[a.isin([a1, a2])], a[a.isin([a1, a2])] == a1)
            self.positivity_estimators_[(a1, a2)] = clf
        return self

    def _create_positivity_profile_columns(self):
        """
        Create a list with the names of the treatment as multi index columns
        Returns:
            a multi index column name
        """
        return pd.MultiIndex.from_tuples(
            self.treatment_pairs_list, names=[self.treatment_a_name, self.treatment_b_name]
        )

    def _treatment_positivity_summary_table(
            self, X: pd.DataFrame, a=None, estimation_population='Both'
    ) -> pd.DataFrame:
        """
            The method summarized the positivity rate per treatment and reference group
        Args:
            X (pd.DataFrame): Covariates to predict on
            a (pd.Series) : The treatment assignment to be passed to the base positivity if required and
            estimation_population (str): Can be treated (treated group only), control (control group only) both
            (either treated or control group) or Every (every patient that was given a treatment). If a is not supplied
            then the mode every is selected automatically.
        Returns:
            pd.DataFrame|pd.Series : a data frame describing the positivity rate of each single positivity estimation

        """
        treatment_profile = self.positivity_profile(X=X, a=a,
                                                    estimation_population=estimation_population).transpose().mean(
            axis=1)
        treatment_profile_df = treatment_profile.reset_index()
        return (
            treatment_profile_df.groupby(
                [self.treatment_a_name, self.treatment_b_name]
            )
                .mean()
                .unstack(-1)
                .droplevel(0, axis=1)
        )

    def _treatment_positivity_summary_contingency(
            self, X: pd.DataFrame, a=None, estimation_population='Both'
    ) -> pd.DataFrame:
        """
        The method returns the positivity profile in a form of contingency matrix where every treatment (be it the one
        or another) appears once in the rows and once in the columns of the profile. This form might be easier for
        viewing in the one versus another case and can be used for graph based algorithms.
        Args:
            X (pd.DataFrame): Covariates to predict on
            a (pd.Series) : The treatment assignment to be passed to the base positivity if required and
            estimation_population (str): Can be treated (treated group only), control (control group only) both
            (either treated or control group) or Every (every patient that was given a treatment) . If a is not supplied
            then the mode every is selected automatically.

        Returns:
            pd.DataFrame: A contingency table of the rate of treatments.

        """
        treatment_profile = self.positivity_profile(X=X, a=a,
                                                    estimation_population=estimation_population).transpose().mean(
            axis=1)
        treats = list(
            {item for sublist in treatment_profile.index for item in sublist}
        )
        treats.sort()
        contingency_frame = pd.DataFrame(index=treats, columns=treats)
        for a1, a2 in combinations(treats, 2):
            if (a1, a2) in treatment_profile.index:
                contingency_frame.loc[a1, a2] = treatment_profile.loc[(a1, a2)]
            else:
                contingency_frame.loc[a1, a2] = treatment_profile.loc[(a2, a1)]
        return contingency_frame.astype(float)

    def _get_treated_rows_indicator(self, X, a, estimation_population='Both', treatment=None):
        """
        Returns indicator variable stating which row in x is to be used for the positivity prediction.
        If the treatment vector is not supplied every coordinate will be used for positivity estimation
        Args:
            X (pd.DataFrame): Covariates
            a (pd.Series) : Treatment assignments of the treated
            estimation_population (str): Can be treated (treated group only), control (control group only) both
            (either treated or control group) or Every (every patient that was given a treatment) . If a is not supplied
            then the mode every is selected automatically.

        Returns:

        """
        if a is None or estimation_population.lower() == 'every':
            return pd.Series(np.repeat(True, X.shape[0]), index=X.index)
        elif estimation_population.lower() == 'both':
            return a.isin(treatment)
        elif estimation_population.lower() == 'control':
            return a == treatment[0]
        elif estimation_population.lower() == 'treated':
            return a == treatment[1]



"""(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Dec 5, 2021
"""
# pylint:disable=protected-access,missing-function-docstring,missing-module-docstring,missing-class-docstring

import unittest

from scipy.special import comb
import numpy as np
import pandas as pd
from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression
from causallib.positivity.multiple_treatment_positivity import (
    OneVersusRestPositivity,
    OneVersusAnotherPositivity,
    )

from causallib.positivity import Trimming, UnivariateBoundingBox


class MultipleTreatmentPositivityTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        X, y = make_classification(
            n_samples=500,
            n_features=3,
            n_informative=2,
            n_redundant=0,
            n_repeated=0,
            n_classes=2,
            n_clusters_per_class=1,
            flip_y=0.01,
            class_sep=1.0,
            random_state=899,
        )
        X, a = X[:, :-1], pd.Series(
            pd.qcut(X[:, -1], 5, labels=[1, 2, 3, 4, 5], retbins=True)[0].to_numpy()
        )
        ab = a % 2
        cls.data = {
            "X": pd.DataFrame(X),
            "a": pd.Series(a),
            "ab": pd.Series(ab),
            "y": pd.Series(y),
        }

    def _get_data_set(self, bin_treat=False):
        if bin_treat:
            return self.data["X"], self.data["ab"], self.data["y"]
        else:
            return self.data["X"], self.data["a"], self.data["y"]

    def test_is_the_one_versus_all_fits(self):
        overlap_calculator = Trimming(LogisticRegression())
        onvsall = OneVersusRestPositivity(overlap_calculator)
        onvsall.fit(self.data["X"], self.data["a"])
        self.assertTrue(
            len(onvsall.positivity_estimators_.keys()) == len(self.data["a"].unique())
        )

    def test_is_the_one_versus_all_fits_transform(self):
        overlap_calculator = Trimming(LogisticRegression())
        onvsall = OneVersusRestPositivity(overlap_calculator)
        a = onvsall.fit_transform(self.data["X"], self.data["a"])
        self.assertEqual(
            len(onvsall.positivity_estimators_.keys()), len(self.data["a"].unique())
        )

    def test_is_the_one_versus_all_predicts(self):
        overlap_calculator = UnivariateBoundingBox(quantile_alpha=0.05)
        onvsall = OneVersusRestPositivity(overlap_calculator)
        onvsall.fit(self.data["X"], self.data["a"])
        pred = onvsall.predict(self.data["X"])
        self.assertTrue(pred.mean() != 1)

    def test_is_the_one_versus_all_positivity_profile(self):
        overlap_calculator = UnivariateBoundingBox(quantile_alpha=0.05)
        onvsall = OneVersusRestPositivity(overlap_calculator)
        onvsall.fit(self.data["X"], self.data["a"])
        pred = onvsall.positivity_profile(self.data["X"])
        svec = [self.data["X"].shape[0], len(self.data["a"].unique())]
        self.assertSequenceEqual(pred.shape, svec)

    def test_is_the_one_versus_all_positivity_profile(self):
        overlap_calculator = UnivariateBoundingBox(quantile_alpha=0.05)
        onvsall = OneVersusRestPositivity(overlap_calculator)
        onvsall.fit(self.data["X"], self.data["a"])
        pred = onvsall._treatment_positivity_summary_contingency(self.data["X"])
        self.assertEqual(pred.shape[0], pred.shape[1])

    def test_is_the_one_versus_all_pred_length(self):
        overlap_calculator = UnivariateBoundingBox(quantile_alpha=0.05)
        onvsall = OneVersusRestPositivity(overlap_calculator)
        onvsall.fit(self.data["X"], self.data["a"])
        pred = onvsall.predict(self.data["X"])
        self.assertTrue(len(pred) == len(self.data["a"]))

    def test_is_the_one_versus_all_pred_values(self):
        overlap_calculator = UnivariateBoundingBox(quantile_alpha=0.2)
        onvsall = OneVersusRestPositivity(overlap_calculator)
        onvsall.fit(self.data["X"], self.data["a"])
        pred = onvsall.predict(self.data["X"])
        np.testing.assert_array_equal(np.sort(np.unique(pred)), [False, True])

    def test_is_the_one_versus_all_pred_values_trimming(self):
        overlap_calculator = Trimming(LogisticRegression())
        onvsall = OneVersusRestPositivity(overlap_calculator)
        onvsall.fit(self.data["X"], self.data["a"])
        pred = onvsall.predict(self.data["X"])
        np.testing.assert_array_equal(np.sort(np.unique(pred)), [False, True])

    def test_is_the_one_versus_all_profile(self):
        overlap_calculator = UnivariateBoundingBox(quantile_alpha=0.2)
        onvsall = OneVersusRestPositivity(overlap_calculator)
        onvsall.fit(self.data["X"], self.data["a"])
        pred = onvsall.positivity_profile(self.data["X"])
        np.testing.assert_array_equal(np.sort(np.unique(pred)[::]), [False, True])

    def test_is_the_one_versus_all_profile_different(self):
        overlap_calculator = UnivariateBoundingBox(quantile_alpha=0.2)
        onvsall = OneVersusRestPositivity(overlap_calculator)
        onvsall.fit(self.data["X"], self.data["a"])
        pred = onvsall.positivity_profile(self.data["X"])
        eq_val = np.sum(pred, axis=1).mean()
        self.assertTrue((eq_val != 0) and (eq_val != len(self.data["a"].unique())))

    def test_is_the_one_versus_all_profile_different_trimming(self):
        overlap_calculator = Trimming(LogisticRegression())
        onvsall = OneVersusRestPositivity(overlap_calculator)
        onvsall.fit(self.data["X"], self.data["a"])
        pred = onvsall.positivity_profile(self.data["X"])
        eq_val = np.sum(pred, axis=1).mean()
        self.assertTrue((eq_val != 0) and (eq_val != len(self.data["a"].unique())))

    def test_logis_one_versus_all(self):
        positivity_estimator = OneVersusRestPositivity(Trimming(LogisticRegression()))
        positivity_estimator.fit(self.data["X"], self.data["a"])
        pred = positivity_estimator.positivity_profile(self.data["X"])
        eq_val = np.sum(pred, axis=1).mean()
        self.assertTrue(eq_val != 0)

    def test_logis_one_versus_all_fit_transform(self):
        positivity_estimator = OneVersusRestPositivity(Trimming(LogisticRegression()))
        pred = positivity_estimator.fit_transform(self.data["X"], self.data["a"])
        pr = positivity_estimator.predict(self.data["X"], self.data["a"])
        self.assertEqual(pred[0].shape[0] ,np.sum(pr))

    def test_logis_one_versus_all_predict_transform(self):
        positivity_estimator = OneVersusRestPositivity(Trimming(LogisticRegression()))
        positivity_estimator.fit(self.data["X"], self.data["a"])
        pred = positivity_estimator.transform(self.data["X"], self.data["a"])
        pr = positivity_estimator.predict(self.data["X"], self.data["a"])
        self.assertEqual(pred[0].shape[0], np.sum(pr))

    def test_one_vs_another_fits(self):
        positivity_estimator = OneVersusAnotherPositivity(
            Trimming(LogisticRegression())
        )
        positivity_estimator.fit(self.data["X"], self.data["a"])
        profile = positivity_estimator.positivity_profile(self.data["X"])
        self.assertEqual(len(profile.columns), comb(len(self.data["a"].unique()), 2))

    def test_one_vs_another_transform(self):
        positivity_estimator = OneVersusAnotherPositivity(
            Trimming(LogisticRegression())
        )
        positivity_estimator.fit(self.data["X"], self.data["a"])
        pos_X, pos_a = positivity_estimator.transform(self.data["X"], self.data["a"])
        self.assertLess(pos_X.shape[0], self.data["X"].shape[0])

    def test_one_vs_another_fit_transform(self):
        positivity_estimator = OneVersusAnotherPositivity(
            Trimming(LogisticRegression())
        )
        res = positivity_estimator.fit_transform(
            self.data["X"], self.data["a"]
        )
        self.assertLess(res[0].shape[0], self.data["X"].shape[0])

    def test_one_vs_another_treat_report(self):
        positivity_estimator = OneVersusAnotherPositivity(
            Trimming(LogisticRegression())
        )
        positivity_estimator.fit(self.data["X"], self.data["a"])
        treatment_rep = positivity_estimator._treatment_positivity_summary_contingency(
            self.data["X"])
        self.assertEqual(len(treatment_rep.columns), len(self.data["a"].unique()))
        self.assertEqual(len(treatment_rep.index), len(self.data["a"].unique()))

    def test_one_vs_another_treat_report_direct(self):
        positivity_estimator = OneVersusAnotherPositivity(
            Trimming(LogisticRegression())
        )
        positivity_estimator.fit(self.data["X"], self.data["a"])
        treatment_rep = positivity_estimator.treatment_positivity_summary(
            self.data["X"], as_contingency_table=True)
        self.assertEqual(len(treatment_rep.columns), len(self.data["a"].unique()))
        self.assertEqual(len(treatment_rep.index), len(self.data["a"].unique()))


    def test_one_vs_another_treat_report_control(self):
        positivity_estimator = OneVersusAnotherPositivity(
            Trimming(LogisticRegression()), treatment_pairs_list=5
        )
        positivity_estimator.fit(self.data["X"], self.data["a"])
        treatment_rep = positivity_estimator.treatment_positivity_summary(
            self.data["X"])
        self.assertEqual(len(treatment_rep.index), len(self.data["a"].unique())-1)


    def test_treatment_population_selection(self):
        positivity_estimator = OneVersusAnotherPositivity(
            Trimming(LogisticRegression()), treatment_pairs_list=5
        )
        positivity_estimator.fit(self.data["X"], self.data["a"])
        profile = positivity_estimator.positivity_profile(self.data["X"])
        self.assertEqual(profile.isna().sum().sum(), 0)

    def test_treatment_population_selection_both(self):
        positivity_estimator = OneVersusAnotherPositivity(
            Trimming(LogisticRegression()))
        positivity_estimator.fit(self.data["X"], self.data["a"])
        profile = positivity_estimator.positivity_profile(self.data["X"], self.data["a"], estimation_population='both')
        for cl in profile.columns:
            self.assertEqual(profile[cl].notna().sum(),  self.data["a"].isin(cl).sum())

    def test_treatment_population_selection_treatment(self):
        positivity_estimator = OneVersusAnotherPositivity(
            Trimming(LogisticRegression()))
        positivity_estimator.fit(self.data["X"], self.data["a"])
        profile = positivity_estimator.positivity_profile(self.data["X"], self.data["a"], estimation_population='treated')
        for cl in profile.columns:
            self.assertEqual(profile[cl].notna().sum(),  (self.data["a"]==cl[0]).sum())


    def test_treatment_population_selection_control(self):
        positivity_estimator = OneVersusAnotherPositivity(
            Trimming(LogisticRegression()))
        positivity_estimator.fit(self.data["X"], self.data["a"])
        profile = positivity_estimator.positivity_profile(self.data["X"], self.data["a"], estimation_population='control')
        for cl in profile.columns:
            self.assertEqual(profile[cl].notna().sum(),  (self.data["a"]==cl[1]).sum())



from __future__ import annotations
import pandas as pd
from causallib.preprocessing.transformers import MatchingTransformer
from causallib.positivity import BasePositivity
from typing import Union, Optional
import sklearn.neighbors
import sklearn.base


class Matching(BasePositivity):

    def __init__(
        self,
        propensity_transform: Optional[sklearn.base.TransformerMixin] = None,
        caliper: Optional[float] = None,
        with_replacement: bool = True,
        n_neighbors: int = 1,
        matching_mode: str = "both",
        metric: str = "mahalanobis",
        knn_backend: Union[str,
                           sklearn.neighbors.NearestNeighbors] = "sklearn",
    ):
        """Fix positivity by matching.

        Args:
            propensity_transform (sklearn.TransformerMixin): an object for data
                preprocessing which implements `fit` and `transform` 
                (default: None)
            caliper (float) : maximal distance for a match to be accepted. If
                not defined, all matches will be accepted. If defined, some
                samples may not be matched and their outcomes will not be
                estimated. (default: None)
            with_replacement (bool): whether samples can be used multiple times
                for matching. If set to False, the matching process will optimize
                the linear sum of distances between pairs of treatment and
                control samples and only `min(N_treatment, N_control)` samples
                will be estimated. Matching with no replacement does not make
                use of the `fit` data and is therefore not implemented for
                out-of-sample data (default: True)
            n_neighbors (int) : number of nearest neighbors to include in match.
                Must be 1 if `with_replacement` is `False.` If larger than 1, the
                estimate is calculated using the `regress_agg_function` or 
                `classify_agg_function` across the `n_neighbors`. Note that when
                the `caliper` variable is set, some samples will have fewer than
                `n_neighbors` matches. (default: 1).
            matching_mode (str) : Direction of matching: `treatment_to_control`,
                `control_to_treatment` or `both` to indicate which set should
                be matched to which. All sets are cross-matched in `match`
                and when `with_replacement` is `False` all matching modes 
                coincide. With replacement there is a difference.
            metric (str) : Distance metric string for calculating distance
                between samples. Note: if an external built `knn_backend`
                object with a different metric is supplied, `metric` needs to
                be changed to reflect that, because `Matching` will set its 
                inverse covariance matrix if "mahalanobis" is set. (default: 
                "mahalanobis", also supported: "euclidean")
            knn_backend (str or callable) : Backend to use for nearest neighbor
                search. Options are "sklearn"  or a callable  which returns an 
                object implementing `fit`, `kneighbors` and `set_params` 
                like the sklearn `NearestNeighbors` object. (default: "sklearn"). 

        """
        self.matching_transformer = MatchingTransformer(
            propensity_transform=propensity_transform,
            caliper=caliper,
            with_replacement=with_replacement,
            n_neighbors=n_neighbors,
            matching_mode=matching_mode,
            metric=metric,
            knn_backend=knn_backend,
        )

    def fit(self, X: pd.DataFrame, a: pd.Series) -> Matching:
        """Fit matching positivity checker.

        Args:
            X (pd.DataFrame): samples
            a (pd.Series): treatment assignment
        """
        self.matching_transformer.fit(X, a, pd.Series())
        return self

    def predict(self, X: pd.DataFrame, a: pd.Series) -> pd.Series:
        """Predict whether or not a sample is in the overlap region.

        Find samples of treatment and control that successfully match and 
        return a boolean indexer which is `True` if they matched and `False` if
        they did not. This function calls the `match` method of the underlying
        `Matching` object.

        Args:
            X (pd.DataFrame): samples
            a (pd.Series): treatment assignment

        Returns:
            pd.Series: a Series of length `X.shape[0]` with the same index as
               `X` and only boolean values
        """
        self.matching_transformer.matching.match(X, a)
        matching_indices = self.matching_transformer.find_indices_of_matched_samples(
            X, a)
        return matching_indices



from .survival_utils import canonize_dtypes_and_names
from .standardized_survival import StandardizedSurvival
from causallib.estimation.base_weight import WeightEstimator
import pandas as pd
from typing import Any, Optional


class WeightedStandardizedSurvival(StandardizedSurvival):
    def __init__(
        self,
        weight_model: WeightEstimator,
        survival_model: Any,
        stratify: bool = True,
        outcome_covariates=None,
        weight_covariates=None,
    ):
        """
        Combines WeightedSurvival and StandardizedSurvival:
            1. Adjusts for treatment assignment by creating weighted pseudo-population (e.g., inverse propensity weighting).
            2. Computes parametric curve by fitting a time-varying hazards model that includes baseline covariates.

        Args:
           weight_model: causallib compatible weight model (e.g., IPW)
           survival_model: Two alternatives:
                1. Scikit-Learn estimator (needs to implement `predict_proba`) - compute parametric curve by fitting a
                    time-varying hazards model that includes baseline covariates. Note that the model is fitted on a
                    person-time table with all covariates, and might be computationally and memory expansive.
                2. lifelines RegressionFitter - use lifelines fitter to compute survival curves from baseline covariates,
                    events and durations
            stratify (bool): if True, fit a separate model per treatment group
            outcome_covariates (array): Covariates to use for outcome model.
                                        If None - all covariates passed will be used.
                                        Either list of column names or boolean mask.
            weight_covariates (array): Covariates to use for weight model.
                                       If None - all covariates passed will be used.
                                       Either list of column names or boolean mask.
        """
        self.weight_model = weight_model
        super().__init__(survival_model=survival_model, stratify=stratify)
        self.outcome_covariates = outcome_covariates
        self.weight_covariates = weight_covariates

    def _prepare_data(self, X, *args, **kwargs):
        """
        Extract the relevant parts for outcome model and weight model for the entire data matrix

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).

        Returns:
            (pd.DataFrame, pd.DataFrame): X_outcome, X_weight
                Data matrix for outcome model and data matrix weight model
        """
        outcome_covariates = X.columns if self.outcome_covariates is None else self.outcome_covariates
        X_outcome = X[outcome_covariates]
        weight_covariates = X.columns if self.weight_covariates is None else self.weight_covariates
        X_weight = X[weight_covariates]
        return X_outcome, X_weight

    def fit(self,
            X: pd.DataFrame,
            a: pd.Series,
            t: pd.Series,
            y: pd.Series,
            w: Optional[pd.Series] = None,
            fit_kwargs: Optional[dict] = None):
        """
        Fits parametric models and calculates internal survival functions.

        Args:
            X (pd.DataFrame): Baseline covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            t (pd.Series): Followup duration, size (num_subjects,).
            y (pd.Series): Observed outcome (1) or right censoring event (0), size (num_subjects,).
            w (pd.Series): NOT USED (for compatibility only) optional subject weights.
            fit_kwargs (dict): Optional kwargs for fit call of survival model

        Returns:
            self
        """
        a, t, y, _, X = canonize_dtypes_and_names(a=a, t=t, y=y, w=None, X=X)
        X_outcome, X_weight = self._prepare_data(X)

        self.weight_model.fit(X=X_weight, a=a, y=y)
        iptw_weights = self.weight_model.compute_weights(X_weight, a)

        # Call fit from StandardizedSurvival, with added ipt weights
        super().fit(X=X_outcome, a=a, t=t, y=y, w=iptw_weights, fit_kwargs=fit_kwargs)
        return self

    def estimate_individual_outcome(
        self,
        X: pd.DataFrame,
        a: pd.Series,
        t: pd.Series,
        y: Optional[Any] = None,
        timeline_start: Optional[int] = None,
        timeline_end: Optional[int] = None
    ) -> pd.DataFrame:
        X_outcome, _ = self._prepare_data(X)
        potential_outcomes = super().estimate_individual_outcome(
            X_outcome,
            a, t, y,
            timeline_start=timeline_start,
            timeline_end=timeline_end,
        )
        return potential_outcomes
